/*
 * Copyright (c) 2001 Navin Boppuri / Prashant Patel
 *	<nboppuri@trinetcommunication.com>,
 *	<pmpatel@trinetcommunication.com>
 * Copyright (c) 2001 Gerd Mennchen <Gerd.Mennchen@icn.siemens.de>
 * Copyright (c) 2001 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * MPC8xx CPM SPI interface.
 *
 * Parts of this code are probably not portable and/or specific to
 * the board which I used for the tests. Please send fixes/complaints
 * to wd@denx.de
 *
 */

#include <ppcboot.h>
#include <mpc8xx.h>
#include <commproc.h>
#include <linux/ctype.h>
#include <malloc.h>

#ifdef	CONFIG_SPI


/* -----------------------------------------------
 * Helper functions to peek into tx and rx buffers
 * ----------------------------------------------- */
static const char * const hex_digit = "0123456789ABCDEF";

static char quickhex (int i)
{
	return hex_digit[i];
}

static void memdump (void *pv, int num)
{
	int i;
	unsigned char *pc = (unsigned char *) pv;

	for (i = 0; i < num; i++)
		printf ("%c%c ", quickhex (pc[i] >> 4), quickhex (pc[i] & 0x0f));
	printf ("\t");
	for (i = 0; i < num; i++)
		printf ("%c", isprint (pc[i]) ? pc[i] : '.');
	printf ("\n");
}

/* -------------------
 * Function prototypes
 * ------------------- */
void spi_init (void);

ssize_t spi_read  (char *, size_t);
ssize_t spi_write (const char *, size_t);

/* -------------------
 * Variables
 * ------------------- */

static ushort r_tbase, r_rbase;
static cbd_t *tbdf, *rbdf;

#define MAX_BUFFER 16

/* RX and TX buffer adresses */
static ulong txbuf;
static ulong rxbuf;

/* **************************************************************************
 *
 *  Function:    spi_init
 *
 *  Description: Init SPI-Controller,
 *               allocate and init buffers and buffer descriptors
 *
 *  return:      ---
 *
 * *********************************************************************** */
void spi_init ()
{
	unsigned int dpaddr;

	volatile spi_t *spi;
	volatile immap_t *immap;
	volatile cpic8xx_t *cpi;
	volatile cpm8xx_t *cp;
	volatile iop8xx_t *iop;

	immap = (immap_t *)  CFG_IMMR;
	cpi   = (cpic8xx_t *)&immap->im_cpic;
	iop   = (iop8xx_t *) &immap->im_ioport;
	cp    = (cpm8xx_t *) &immap->im_cpm;

	spi   = (spi_t *)&cp->cp_dparam[PROFF_SPI];

/* 1 */
	/* ------------------------------------------------
	 * Initialize Port B SPI pins -> page 34-8 MPC860UM
	 * (we are only in Master Mode !)
	 * ------------------------------------------------ */

	/* --------------------------------------------
	 * GPIO or per. Function
	 * PBPAR[28] = 1 [0x00000008] -> PERI: (SPIMISO)
	 * PBPAR[29] = 1 [0x00000004] -> PERI: (SPIMOSI)
	 * PBPAR[30] = 1 [0x00000002] -> PERI: (SPICLK)
	 * PBPAR[31] = 0 [0x00000001] -> GPIO: (CS for PCUE-EEPROM)
	 * -------------------------------------------- */
	cp->cp_pbpar |=  0x0000000E;	/* set  bits	*/
	cp->cp_pbpar &= ~0x00000001;	/* reset bit	*/

	printf("*** SPI Init: PBPAR  ADR=0x%08x  VAL=0x%08x\n",
		(unsigned int)&cp->cp_pbpar, cp->cp_pbpar);

	/* ----------------------------------------------
	 * In/Out or per. Function 0/1
	 * PBDIR[28] = 1 [0x00000008] -> PERI1: SPIMISO
	 * PBDIR[29] = 1 [0x00000004] -> PERI1: SPIMOSI
	 * PBDIR[30] = 1 [0x00000002] -> PERI1: SPICLK
	 * PBDIR[31] = 1 [0x00000001] -> GPIO OUT: CS for PCUE-EEPROM
	 * ---------------------------------------------- */
	cp->cp_pbdir |= 0x0000000F;

	printf("*** SPI Init: PBDIR  ADR=0x%08x  VAL=0x%08x\n",
		(unsigned int)&cp->cp_pbdir, cp->cp_pbdir);

	/* ----------------------------------------------
	 * open drain or active output
	 * PBODR[28] = 1 [0x00000008] -> open drain: SPIMISO
	 * PBODR[29] = 0 [0x00000004] -> active output SPIMOSI
	 * PBODR[30] = 0 [0x00000002] -> active output: SPICLK
	 * PBODR[31] = 0 [0x00000001] -> active output: GPIO OUT: CS for PCUE-EEPROM
	 * ---------------------------------------------- */

	cp->cp_pbodr |=  0x00000008;
	cp->cp_pbodr &= ~0x00000007;

	printf("*** SPI Init: PBODR  ADR=0x%08x  VAL=0x%04x\n",
		(unsigned int)&cp->cp_pbodr, cp->cp_pbodr);

	printf ("*** SPI Init: spi = 0x%08x ***\n", (unsigned int) spi);

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero
	 */
	spi->spi_rstate	= 0;
	spi->spi_rdp	= 0;
	spi->spi_rbptr	= 0;
	spi->spi_rbc	= 0;
	spi->spi_rxtmp	= 0;
	spi->spi_tstate	= 0;
	spi->spi_tdp	= 0;
	spi->spi_tbptr	= 0;
	spi->spi_tbc	= 0;
	spi->spi_txtmp	= 0;

	/* Allocate space for one transmit and one receive buffer
	 * descriptor in the DP ram
	 */
#ifdef CFG_ALLOC_DPRAM
	dpaddr = dpram_alloc_align (sizeof(cbd_t)*2, 8);
#else
	dpaddr = CPM_SPI_BASE;
#endif

	printf ("*** SPI Init: dpaddr = 0x%08x ***\n",
			(unsigned int) dpaddr);

/* 3 */
	/* Set up the SPI parameters in the parameter ram */
	spi->spi_rbase = r_rbase = dpaddr;
	spi->spi_tbase = r_tbase = dpaddr + sizeof (cbd_t);

	/***********IMPORTANT******************/

        /*
         * Setting transmit and receive buffer descriptor pointers
         * initially to rbase and tbase. Only the microcode patches
         * documentation talks about initializing this pointer. This
         * is missing from the sample I2C driver. If you dont
         * initialize these pointers, the kernel hangs.
	 */
	spi->spi_rbptr = spi->spi_rbase;
	spi->spi_tbptr = spi->spi_tbase;

/* 4 */
	/* Init SPI Tx + Rx Parameters */
	while (cp->cp_cpcr & CPM_CR_FLG)
		;
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SPI, CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG)
		;

/* 5 */
	/* Set SDMA configuration register */
	immap->im_siu_conf.sc_sdcr = 0x0001;

/* 6 */
	/* Set to big endian. */
	spi->spi_tfcr = SMC_EB;
	spi->spi_rfcr = SMC_EB;

/* 7 */
	/* Set maximum receive size. */
	spi->spi_mrblr = 64;

/* 8 + 9 */
	/* tx and rx buffer descriptors */
	tbdf = (cbd_t *) & cp->cp_dpmem[r_tbase];
	rbdf = (cbd_t *) & cp->cp_dpmem[r_rbase];

	/* initialize tx and tx bd's */
	tbdf->cbd_sc &= ~BD_SC_READY;
	rbdf->cbd_sc &= ~BD_SC_EMPTY;

	/* Allocate memory for RX and TX buffers */
	rxbuf = (ulong) malloc (MAX_BUFFER);
	txbuf = (ulong) malloc (MAX_BUFFER);
	printf ("*** SPI Init: rxbuf = 0x%08lx ***\n", rxbuf);
	printf ("*** SPI Init: txbuf = 0x%08lx ***\n", txbuf);

	/* Set the bd's rx and tx buffer address pointers */
	rbdf->cbd_bufaddr = rxbuf;
	tbdf->cbd_bufaddr = txbuf;
	printf ("*** SPI Init: rbdf->cbd_bufaddr = 0x%08x ***\n",
			rbdf->cbd_bufaddr);
	printf ("*** SPI Init: tbdf->cbd_bufaddr = 0x%08x ***\n",
			tbdf->cbd_bufaddr);

/* 10 + 11 */
	cp->cp_spim = 0;			/* Mask  all SPI events */
	cp->cp_spie = SPI_EMASK;		/* Clear all SPI events	*/

	return;
}

#if 0
	/* Shut down SPI */
	cp->cp_spmode = 0;
	cp->cp_spie = 0xff;
	cp->cp_spim = 0;
#endif

/****************************************************************************
 *  spi_write
 **************************************************************************** */
ssize_t spi_write (const char *buf, size_t count)
{
	volatile immap_t *immap;
	volatile cpm8xx_t *cp;
	int tm;

	printf("*** spi_write entered ***\n");

	if (count > MAX_BUFFER) {
printf ("Count %d > MAX_BUFFER %d --> truncating\n", count, MAX_BUFFER);
		count = MAX_BUFFER;
	}

	immap = (immap_t *) CFG_IMMR;	/* and to internal registers */
	cp    = (cpm8xx_t *) &immap->im_cpm;

	/* Initialize buffer memory to empty */
	memset ((void *) txbuf, 0, MAX_BUFFER);
	memset ((void *) rxbuf, 0, MAX_BUFFER);

	/* copy transmit data from user space to tx buffer */
	memcpy ((void *) txbuf, (void *) buf, count);
printf ("\nspi_write: txbuf initialized with user space buffer\n");
memdump ((void *) txbuf, MAX_BUFFER); /* dump of txbuf before tx */
printf ("spi_write: rxbuf zeroed\n");
memdump ((void *) rxbuf, MAX_BUFFER); /* dump of rxbuf before tx */
printf("\n");

	/* Setting tx bd status and data length */
	tbdf->cbd_sc  = BD_SC_READY | BD_SC_LAST | BD_SC_WRAP;
	/* tbdf->cbd_sc |= BD_SC_INTRPT; */
	tbdf->cbd_datlen = count;

	printf("*** spi_write: Bytes to be written: %d ***\n", tbdf->cbd_datlen);

	/* Setting rx bd status and data length */
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP;
	rbdf->cbd_datlen = 0;	 /* rx length = 0 since we are not reading here */

	cp->cp_spmode = SPMODE_DIV16	|
			SPMODE_REV	|
			SPMODE_MSTR	|
			SPMODE_EN	|
			SPMODE_LEN(8)	|	/* 8 Bits per char */
			SPMODE_PM(0x0F) ;
	cp->cp_spim = 0;			/* Mask  all SPI events */
	cp->cp_spie = SPI_EMASK;		/* Clear all SPI events	*/

	/* start spi transfer */
	printf ("*** spi_write: Performing transfer ...\n");
	cp->cp_spcom |= SPI_STR;		/* Start transmit */

	/* --------------------------------
	 * Wait for SPI transmit to get out
	 * or time out (1 second = 1000 ms)
	 * -------------------------------- */
	for (tm=0; tm<1000; ++tm) {
		if (cp->cp_spie & SPI_TXB) {	/* Tx Buffer Empty */
printf("*** spi_write: Tx buffer empty\n");
			break;
		}
		if ((tbdf->cbd_sc & BD_SC_READY) == 0) {
printf("*** spi_write: Tx BD done\n");
			break;
		}
		udelay (1000);
	}
	if (tm >= 1000) {
		printf ("*** spi_write: Time out while writing to SPI!\n");
	}
	printf ("*** spi_write: ... transfer ended\n");

	/* Ignore received data */
	printf ("\nspi_write: txbuf after sending\n");
	memdump ((void *) txbuf, MAX_BUFFER);
	printf ("spi_write: received rxbuf in spi_write -> ignore\n");
	memdump ((void *) rxbuf, MAX_BUFFER);
	printf ("\n");

	return count;
}


/****************************************************************************
 *  Function:    spi_read
 **************************************************************************** */
ssize_t spi_read (char *buf, size_t count)
{
	volatile immap_t *immap;
	volatile cpm8xx_t *cp;
	int tm;

	printf("*** spi_read entered ***\n");

	if (count > MAX_BUFFER) {
printf ("Count %d > MAX_BUFFER %d --> truncating\n", count, MAX_BUFFER);
		count = MAX_BUFFER;
	}

	immap = (immap_t *) CFG_IMMR;	/* and to internal registers */
	cp    = (cpm8xx_t *) &immap->im_cpm;

	/* Initialize buffer memory to empty */
	memset ((void *) txbuf, 0, MAX_BUFFER);
	memset ((void *) rxbuf, 0, MAX_BUFFER);

	/* copy transmit data from user space to tx buffer */
	memcpy ((void *) txbuf, (void *) buf, count);
printf ("\nspi_read: txbuf zeroed\n");
	memdump ((void *) txbuf, MAX_BUFFER);	/* dump of txbuf before transmit */
printf ("spi_read: rxbuf undef (zeroed)\n");
	memdump ((void *) rxbuf, MAX_BUFFER);	/* dump of rxbuf before transmit */
	printf ("\n");

	/* Setting tx bd status and data length */
	tbdf->cbd_datlen = count;
	tbdf->cbd_sc  = BD_SC_LAST | BD_SC_WRAP;
	tbdf->cbd_sc |= BD_SC_READY;

	/* Setting rx bd status and data length */
	rbdf->cbd_datlen = count;
	rbdf->cbd_sc  = BD_SC_EMPTY | BD_SC_WRAP;

	printf ("*** TX BD: "); memdump ((void *) tbdf, sizeof(cbd_t));
	printf ("*** RX BD: "); memdump ((void *) rbdf, sizeof(cbd_t));
	printf ("\n");

	cp->cp_spmode = SPMODE_DIV16	|
			SPMODE_REV	|
			SPMODE_MSTR	|
			SPMODE_EN	|
			SPMODE_LEN(8)	|	/* 8 Bits per char */
			SPMODE_PM(0x0F) ;
	cp->cp_spim = 0;			/* Mask  all SPI events */
	cp->cp_spie = SPI_EMASK;		/* Clear all SPI events	*/

	/* start spi transfer */
	printf ("*** spi_read: Performing transfer ...\n");

	cp->cp_spcom |= SPI_STR;

	/* --------------------------------
	 * Wait for SPI receive to complete
	 * or time out (1 second = 1000 ms)
	 * -------------------------------- */
	for (tm=0; tm<1000; ++tm) {
		if (cp->cp_spie & SPI_RXB) {	/* Rx Buffer closed */
printf("*** spi_read: Rx buffer closed\n");
			break;
		}
		if ((rbdf->cbd_sc & BD_SC_EMPTY) == 0) {
printf("*** spi_read: Rx BD done\n");
			break;
		}
		udelay (1000);
	}
	if (tm >= 1000) {
		printf ("*** spi_read: Time out while reading from SPI!\n");
	}
	printf ("*** spi_read: ... transfer ended\n");

	printf ("*** SPI EVENTS: %04X\n", cp->cp_spie);

	printf ("*** TX BD: "); memdump ((void *) tbdf, sizeof(cbd_t));
	printf ("*** RX BD: "); memdump ((void *) rbdf, sizeof(cbd_t));
	printf ("\n");

if ((rbdf->cbd_sc & BD_SC_EMPTY) == 0 && (rbdf->cbd_sc & BD_SC_LAST)) {
  printf ("*** Buffer is closed\n");
}
	printf ("\nspi_read: txbuf after read\n");
	memdump ((void *) txbuf, MAX_BUFFER);	/* dump of txbuf before transmit */
	printf ("spi_read: rxbuf after read\n");
	memdump ((void *) rxbuf, MAX_BUFFER);	/* dump of rxbuf after transmit */
	printf ("\n");

	memcpy ((void *) buf, (void *) rxbuf, count);

	return count;
}

#endif	/* CONFIG_SPI */
